# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import platform
import shutil
import sys

from github import Github
from semantic_version import SimpleSpec, Version

from .cmd_common import WorkflowTask, GH_USER_AGENT

from .util import (
    check_arg,
    log_info,
    log_warning,
    to_list,
    write,
)


class CheckTask(WorkflowTask):
    DEFAULT_OPTS = {
        "branches": None,
        "can_push": None,
        "clean": None,
        "github": None,
        "os": None,
        "python": None,
        "twine": None,
        "up_to_date": None,
        "venv": None,
        "version": None,
    }

    def __init__(self, config):
        super().__init__(config)
        opts = self.opts

        check_arg(opts["branches"], (str, list, tuple), or_none=True)
        check_arg(opts["can_push"], bool, or_none=True)
        check_arg(opts["clean"], bool, or_none=True)
        check_arg(opts["os"], (str, list, tuple), or_none=True)
        check_arg(opts["python"], (str, SimpleSpec), or_none=True)
        check_arg(opts["twine"], bool, or_none=True)
        check_arg(opts["up_to_date"], bool, or_none=True)
        check_arg(opts["venv"], bool, or_none=True)
        # check_arg(opts["gh_auth"], dict, or_none=True)

        opts["branches"] = to_list(opts["branches"])
        opts["os"] = to_list(opts["os"])
        if isinstance(opts["python"], str):
            opts["python"] = SimpleSpec(opts["python"])

    def to_str(self, context):
        checks = ", ".join([k for k, v in sorted(self.opts.items()) if v])
        return "{}({})".format(self.__class__.__name__, checks)

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        # Additional arguments for the 'run' command
        run_parser.add_argument(
            "--no-check",
            action="store_true",
            help="don't let the 'check' task stop the workflow (log warnings instead)",
        )

        sp = subparsers.add_parser(
            "check", parents=parents, help="check preconditions",
        )

        sp.add_argument(
            "--branch",
            nargs="*",
            metavar="BRANCH",
            dest="branches",
            help="check if current branch is allowed",
        )
        sp.add_argument(
            "--can-push", action="store_true", help="assert `git push` would work",
        )
        sp.add_argument(
            "--venv",
            action="store_true",
            help="assert running inside a virtual environment",
        )
        sp.add_argument(
            "--github", action="store_true", help="assert gitub repo is available",
        )
        sp.add_argument(
            "--twine", action="store_true", help="assert twine is available",
        )
        sp.add_argument(
            "--clean",
            action="store_true",
            help="assert that all changes are committed",
        )
        sp.add_argument(
            "--os",
            choices=["Linux", "Darwin", "Windows"],
            nargs="*",
            help="check current operating system",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        opts = self.opts
        FLAG_ERROR = 1024
        FLAG_UP_TO_DATE = 512

        err_list = []

        def _ok(msg, output=None):
            write(msg, "info", "check", output)

        def _error(msg, output=None):
            write(msg, "error", "check", output)
            err_list.append(msg)

        def _warn(msg, output=None):
            write(msg, "warning", "check", output)

        # project_info = get_project_info()
        # print(project_info)

        repo = context.repo_obj
        remote = repo.remote()

        # dist = pkg_resources.get_distribution("yabs")
        # assert dist.version == __version__
        # print(opts)
        if opts["branches"]:
            cur_branch = repo.active_branch.name
            if cur_branch in opts["branches"]:
                _ok(
                    "Active branch {!r} is in allowed list ({}).".format(
                        cur_branch, ", ".join(opts["branches"])
                    )
                )
            else:
                _error(
                    "Active branch {!r} not in allowed list ({}).".format(
                        cur_branch, ", ".join(opts["branches"])
                    )
                )

        if opts["can_push"]:
            info = remote.push(dry_run=True)[0]

            if info.flags & FLAG_ERROR:
                msg = "`git push` would fail (flags: {})".format(info.flags)
                _error(msg, info.summary)
            elif not (info.flags & FLAG_UP_TO_DATE):
                msg = "`git push` would transfer data (flags: {})".format(info.flags)
                _warn(msg, info.summary)
            else:
                _ok("`git push` would succeed.")

        # if opts["cmp_version"]:
        #     raise NotImplementedError

        # if opts["connected"]:
        #     raise NotImplementedError

        if opts["clean"]:
            if repo.is_dirty():
                msg = "Repository has pending commits"
                _error(msg, repo.git.status())
            else:
                _ok("Repository is clean.")

        if opts["github"]:
            repo_name = opts.get("repo") or context.repo
            if not repo_name or "/" not in repo_name:
                _error(
                    "Invalid repo name (expected `GH-USER/PROJECT`): {}".format(
                        repo_name
                    )
                )
            token = context.gh_auth_token
            try:
                gh = Github(token, user_agent=GH_USER_AGENT)
                gh_repo = gh.get_repo(repo_name, lazy=False)
                full_name = gh_repo.full_name
                _ok("GitHub repo {} is accessible: {}".format(repo_name, full_name))
            except Exception as e:
                _error("Could not access GitHub repo {}: {}".format(repo_name, e))

        if opts["os"]:
            info = remote.push(dry_run=True)[0]
            # print("push", info.flags, info.summary)
            ps = platform.system()
            if ps in opts["os"]:
                _ok(
                    "Platform {!r} is in allowed list ({}).".format(
                        ps, ", ".join(opts["os"])
                    )
                )
            else:
                _error(
                    "Platform {!r} not in allowed list ({}).".format(
                        ps, ", ".join(opts["os"])
                    )
                )

        if opts["python"]:
            py_ver = Version(".".join(map(str, sys.version_info[:3])))
            if opts["python"].match(py_ver):
                _ok("Python version {} matches {}.".format(py_ver, opts["python"]))
            else:
                _error(
                    "Python version {} does not match {}.".format(
                        py_ver, opts["python"]
                    )
                )

        if opts["twine"]:
            if shutil.which("twine"):
                _ok("`twine` is available.")
            else:
                _error("`twine` not available.")

        if opts["up_to_date"]:
            remote.update()
            status = repo.git.status("-uno", porcelain=False)
            if 'use "git pull"' in status:
                msg = "Remote branch contains unpulled changes"
                _error(msg, status)
            else:
                _ok("Remote branch has not diverged.")

        if opts["venv"]:
            is_venv = hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix
            if is_venv:
                _ok("Running inside a virtual environment.")
            else:
                _error("Not running inside a virtual environment.")

        if opts["version"]:
            _ret_code, real_version = self._exec(
                ["python", "setup.py", "--version"], quiet=True
            )
            vm = context.version_manager
            if real_version != str(vm.master_version):
                _error(
                    "`setup.py --version` returned {} (expected {}).".format(
                        real_version, vm.master_version
                    )
                )
            else:
                _ok("`setup.py --version` returned {}.".format(real_version))

        if err_list:
            write("Checks failed:\n  - {}".format("\n  - ".join(err_list)), "error")
            if context._args.no_check:
                log_warning("`--no-check` was passed: ignoring the errors above.")
            else:
                log_info("Use the `--no-check` argument to ignore the errors above.")
                return False
        return True



# -*- coding: utf-8 -*-
"""
A collection of tools for tox release workflows.

(c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php

Usage examples:
    $ yabs --help
    $ yabs check --clean --branch master
"""
import argparse
import platform
import sys

from snazzy import enable_colors
from yabs import __version__

from .cmd_common import register_cli_commands
from .task_runner import handle_run_command
from .util import init_logging

# --- verbose_parser ----------------------------------------------------------

verbose_parser = argparse.ArgumentParser(
    add_help=False,
    # allow_abbrev=False,
)

qv_group = verbose_parser.add_mutually_exclusive_group()
qv_group.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=3,
    help="increment verbosity by one (default: %(default)s, range: 0..5)",
)
qv_group.add_argument(
    "-q", "--quiet", default=0, action="count", help="decrement verbosity by one"
)

# --- common_parser ----------------------------------------------------------

common_parser = argparse.ArgumentParser(
    add_help=False,
    # allow_abbrev=False,
)
common_parser.add_argument(
    "-n",
    "--dry-run",
    action="store_true",
    help="just simulate and log results, but don't change anything",
)
common_parser.add_argument(
    "--no-color", action="store_true", help="prevent use of ansi terminal color codes"
)


# ===============================================================================
# run
# ===============================================================================
def run():
    """CLI main entry point."""

    parents = [verbose_parser, common_parser]

    parser = argparse.ArgumentParser(
        description="Release workflow automation tools.",
        epilog="See also https://github.com/mar10/yabs",
        parents=parents,
        # allow_abbrev=False,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="display version info and exit (combine with -v for more information)",
    )
    subparsers = parser.add_subparsers(help="sub-command help")

    # --- Create the parser for the "run" command -----------------------------

    sp = subparsers.add_parser(
        "run", parents=parents, help="run a workflow definition",
    )
    sp.add_argument(
        "workflow", nargs="?", default="./yabs.yaml", help="run a workflow definition",
    )
    sp.add_argument(
        "--force",
        action="store_true",
        help="allow to ignore some errors (like bumping above `max_increment`)",
    )
    sp.add_argument(
        "--no-release",
        action="store_true",
        help="don't upload tags and assets to GitHub or PyPI (but still build assets)",
    )
    sp.set_defaults(command=handle_run_command)
    run_parser = sp

    # --- Let all sublasses of `WorkflowTask` add their arguments --------------

    register_cli_commands(subparsers, parents, run_parser)

    # --- Parse command line ---------------------------------------------------

    args = parser.parse_args()

    args.verbose -= args.quiet
    del args.quiet

    # print("verbose", args.verbose)
    init_logging(args.verbose)  # , args.log_file)

    if not args.no_color:
        # Enable terminal colors (if sys.stdout.isatty())
        enable_colors(True, force=False)

    if getattr(args, "version", None):
        if args.verbose >= 4:
            PYTHON_VERSION = "{}.{}.{}".format(
                sys.version_info[0], sys.version_info[1], sys.version_info[2]
            )
            version_info = "yabs/{} Python/{}({} bit) {}".format(
                __version__,
                PYTHON_VERSION,
                "64" if sys.maxsize > 2 ** 32 else "32",
                platform.platform(),
            )
            version_info += "\nPython from: {}".format(sys.executable)
        else:
            version_info = __version__
        print(version_info)
        sys.exit(0)

    if not callable(getattr(args, "command", None)):
        parser.error("missing command")

    try:
        return args.command(parser, args)
    except KeyboardInterrupt:
        print("\nAborted by user.", file=sys.stderr)
        sys.exit(3)
    # Unreachable...
    return


# Script entry point
if __name__ == "__main__":
    # Just in case...
    from multiprocessing import freeze_support

    freeze_support()

    run()



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os

from git import Repo
from git.exc import GitCommandError

from .cmd_common import WorkflowTask
from .util import check_arg, log_dry, log_response


class TagTask(WorkflowTask):
    DEFAULT_OPTS = {
        "name": "v{version}",
        "message": "Version {version}",
    }

    def __init__(self, opts):
        super().__init__(opts)

        opts = self.opts
        check_arg(opts["name"], str)
        check_arg(opts["message"], str)

    # def to_str(self, context):
    #     add = self.opts["add"] or self.opts["add_known"]
    #     return "{}(add: {}, '{}')".format(
    #         self.__class__.__name__, add, self.opts["message"]
    #     )

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        """"""
        sp = subparsers.add_parser(
            "tag",
            parents=parents,
            help="increment current 'patch' version (add `--minor` or `--major`)",
        )
        sp.add_argument(
            "name", help="tag name",
        )
        sp.add_argument(
            "message", help="tag message",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        opts = self.opts
        name = opts["name"].format(**vars(context))
        message = opts["message"].format(**vars(context))

        repo_path = os.path.abspath(".")
        repo = Repo(repo_path)
        git = repo.git

        if self.dry_run:
            log_dry("git tag -a {}".format(name))
            context.tag_name = name
            return True
        try:
            res = git.tag(
                name,
                annotate=True,
                message=message,
                dry_run=self.dry_run,
                verbose=self.verbose >= 4,
            )
            log_response("git tag {}".format(name), res, "info", self.dry_run)
            context.tag_name = name
        except GitCommandError as e:
            log_response("git tag", "{}".format(e), "error", self.dry_run)
            return False
        return True



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
from pathlib import Path

from .cmd_common import WorkflowTask
from .util import (
    ConfigError,
    check_arg,
    log_dry,
    log_error,
    log_info,
    log_ok,
    log_warning,
    logger,
    remove_directory,
)


def get_folder_file_names(folder):
    """Return folder files names as set."""
    p = Path(folder)
    name_set = set([e.name for e in p.iterdir()])
    return name_set


class PypiReleaseTask(WorkflowTask):
    KNOWN_TARGETS = ("sdist", "bdist_wheel", "bdist_msi")
    DEFAULT_OPTS = {
        "build": ["sdist", "bdist_wheel"],
        "clean": True,
        "ignore_errors": False,
        "revert_bump_on_error": True,
        "silent": False,
        "upload": False,
    }

    def __init__(self, opts):
        super().__init__(opts)

        opts = self.opts
        check_arg(opts["clean"], bool)
        check_arg(opts["build"], list)
        unknown = set(opts["build"]).difference(set(self.KNOWN_TARGETS))
        if unknown:
            raise ConfigError(
                "Unkown `pypi_release.build` value: {}".format(", ".join(unknown))
            )
        check_arg(opts["upload"], bool)

    def to_str(self, context):
        opts = self.opts
        args = "{}".format(", ".join(opts["build"]))
        if opts["upload"]:
            args += " & upload"
        return "{}(build {})".format(self.__class__.__name__, args)

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        """"""
        sp = subparsers.add_parser(
            "pypi-release",
            parents=parents,
            help="Make sdist, wheel, and upload on PyPI",
        )
        sp.add_argument(
            "--clean", action="store_true", help="erase dist folder before building",
        )
        sp.add_argument(
            "--upload",
            action="store_true",
            help="upload built artefacts to PyPI using twine",
        )
        sp.add_argument(
            "--build",
            choices=cls.KNOWN_TARGETS,
            nargs="*",
            help="call setup.py to build artefacts (option may be repeated, default: sdist, bdist_wheel)",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    def _run(self, context):
        opts = self.opts
        ok = True

        extra_args = []
        # NOTE: `--dry-run` flag does not work well with setup.py?
        #    Seems to produce errors like
        #    "error: [Errno 2] No such file or directory: 'test-release-tool-0.0.1/PKG-INFO'"
        # if self.dry_run:
        #     extra_args.append("--dry-run")

        if self.verbose >= 4:
            extra_args.append("--verbose")
        elif self.verbose <= 2:
            extra_args.append("--quiet")

        # Check if setup.py really uses the expected name & version
        ret_code, real_name = self._exec(["python", "setup.py", "--name"] + extra_args)
        ret_code, real_version = self._exec(
            ["python", "setup.py", "--version"] + extra_args
        )

        if real_version != str(context.version):
            if not self.dry_run:
                raise RuntimeError(
                    "`setup.py --version` returned {} (expected {})".format(
                        real_version, context.version
                    )
                )

        # It's hard to guess the resulting name of the created artefacts,
        # so if don't want to erase the target '/dist' folder, we need to
        #   1. create / empty a temp folder
        #   2. build into that temp folder
        #   3. twine upload that folder
        #   4. move files from temp to /dist (overwrite if neccessary
        #   5. remove temp folder
        # org_dist_dir = os.path.abspath("dist")
        org_dist_dir = Path("dist").absolute().resolve()
        if not org_dist_dir.is_dir():
            raise RuntimeError("Folder not found: {}".format(org_dist_dir))

        temp_dist_dir = Path("dist.yabs").absolute().resolve()
        if temp_dist_dir.exists():
            remove_directory(temp_dist_dir, content_only=True, log=logger.info)
        else:
            temp_dist_dir.mkdir()

        extra_args
        # dist_files = []
        # dist_prefix = "{}-{}".format(real_name, real_version)

        for target in self.opts["build"]:
            log_info("Building {} for {} {}...".format(target, real_name, real_version))
            prev_artefacts = get_folder_file_names(temp_dist_dir)
            ret_code, _out = self._exec(
                ["python", "setup.py", target, "--dist-dir", str(temp_dist_dir)]
                + extra_args
            )
            new_artefacts = list(
                get_folder_file_names(temp_dist_dir).difference(prev_artefacts)
            )
            # print("1", prev_artefacts)
            # print("2", new_artefacts)
            ok = ok and (ret_code == 0)
            if len(new_artefacts) != 1:
                raise RuntimeError(
                    "Created {} artefacts (expected 1): {}".format(
                        len(new_artefacts), new_artefacts
                    )
                )
            artefact = new_artefacts[0]

            context.artefacts[target] = artefact

            if ret_code == 0:
                log_ok(
                    "Created '{}': {} {}: {}".format(
                        target, real_name, real_version, artefact
                    )
                )
            else:
                log_error(
                    "Failed to build '{}': {} {} {}".format(
                        target, real_name, real_version, ", ".join(new_artefacts)
                    )
                )

        twine_pattern = "{}/*".format(temp_dist_dir)
        ret_code, _out = self._exec(["twine", "check", twine_pattern])
        ok = ok and (ret_code == 0)

        if opts["upload"]:
            if context._args.no_release:
                log_warning(
                    "`--no-release` was passed: skipping `twine upload` of 'pypi_release' task."
                )
            elif self.dry_run:
                log_dry("twine upload {}".format(twine_pattern))
            else:
                # TODO: uses .pypirc
                ret_code, _out = self._exec(
                    [
                        "twine",
                        "upload",
                        "--non-interactive",
                        "--verbose",
                        # "--skip-existing",
                        "--disable-progress-bar",
                        twine_pattern,
                    ]
                )
                ok = ok and (ret_code == 0)

        for src in temp_dist_dir.iterdir():
            if self.dry_run:
                log_dry("Move {} => {}".format(src, org_dist_dir.joinpath(src.name)))
            else:
                log_info("Move {} => {}".format(src, org_dist_dir.joinpath(src.name)))
                src.replace(org_dist_dir.joinpath(src.name))
        # Remove the temp folder (even in dry-run), since we don't want to
        # commit it:
        remove_directory(temp_dist_dir, log=logger.info)
        # Adjust artefact paths:
        d = {}
        for target, path in context.artefacts.items():
            path_new = org_dist_dir.joinpath(Path(path).name)
            if not path_new.is_file() and not self.dry_run:
                raise RuntimeError("Artefact not found {}".format(path_new))
            d[target] = path_new
        context.artefacts = d
        # print("np", context.artefacts)

        if opts["clean"]:
            ret_code, _out = self._exec(
                ["python", "setup.py", "clean", "--all"] + extra_args
            )
            ok = ok and (ret_code == 0)

        return ok

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        try:
            res = self._run(context)
            if res:
                return res
        except Exception as e:
            log_error("{}".format(e))

        if self.opts["revert_bump_on_error"] and not self.dry_run:
            # log_warning("Reverting bump {} => {} ...".format(vm.master_version, context.version))
            vm = context.version_manager
            vm.reset_version(write=not self.dry_run)
            context.version = vm.master_version

        return False



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os

from git import Repo

from .cmd_common import WorkflowTask
from .util import check_arg, log_dry, log_response


class CommitTask(WorkflowTask):
    DEFAULT_OPTS = {
        "add": [],  # Also `git add` these files ('.' for all)
        "add_known": True,  # Commit with -a flag
        "message": "Bumping version to {version}",
    }

    def __init__(self, opts):
        super().__init__(opts)

        opts = self.opts
        check_arg(opts["add"], (list, tuple))
        check_arg(opts["add_known"], bool)
        check_arg(opts["message"], str)

    def to_str(self, context):
        opts = self.opts
        add = opts["add"] or opts["add_known"]
        message = opts["message"].format(**vars(context))
        return "{}(add: {}, '{}')".format(self.__class__.__name__, add, message.strip())

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        """"""
        sp = subparsers.add_parser(
            "commit",
            parents=parents,
            help="increment current 'patch' version (add `--minor` or `--major`)",
        )
        sp.add_argument(
            "--add-known", action="store_true", help="",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        opts = self.opts
        message = opts["message"].format(**vars(context))

        repo_path = os.path.abspath(".")
        repo = Repo(repo_path)
        git = repo.git
        # remote = repo.remote()
        index = repo.index

        opts_add = opts["add"]
        opts_add_known = opts["add_known"]

        if opts_add:
            check_arg(opts_add, (list, tuple))
            index.add(opts_add, write=not self.dry_run)

        if opts_add_known:
            res = git.add(".", dry_run=self.dry_run, verbose=self.verbose >= 4)
            log_response("git add .", res, "debug", self.dry_run)

        if self.dry_run:
            log_dry("commit")
        else:
            # if opts["add_known"]:
            index.commit(message,)
        return True



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os
import re
import tempfile
from abc import ABC, abstractmethod
from configparser import ConfigParser

import toml
from semantic_version import Version

from .util import check_arg, log_debug, log_info, log_warning, resolve_path

ORDERED_INCREMENTS = ("postrelease", "prerelease", "patch", "minor", "major")
INCREMENTS = frozenset(ORDERED_INCREMENTS)


def copy_version(v, prerelease=None):
    return Version(major=v.major, minor=v.minor, patch=v.patch, prerelease=prerelease)


class SafeFileWriter:
    """Open a file for writing, create a backup and prevent corruption on errors.

    Example:
        with FileWriter(fspec, "wt") as f:
            # on enter:
            #   - Remove fspec.tmp if any
            #   - self.f = open(fspec.tmp, mode)

            f.write(...)

            # on exit:
            #   - Remove fspec.bak if any
            #   - Move fspec -> fspec.bak if fspec exists
            #   - Move fspec.tmp -> fspec
            # on error:
            #   - Close and remove fspec.tmp
    """

    backup_ext = "bak"
    temp_ext = "tmp"

    def __init__(
        self,
        fspec,
        mode,
        keep_backup=True,
        use_peer_tmp=True,
        keep_temp_on_error=False,
    ):
        if "w" not in mode:
            raise RuntimeError("`mode` must be writable")
        self.fspec = fspec
        self.mode = mode
        self.keep_backup = keep_backup
        self.use_peer_tmp = use_peer_tmp
        self.keep_temp_on_error = keep_temp_on_error
        self.f = None
        self.temp_file = None

    def __enter__(self):
        if self.use_peer_tmp:
            self.temp_file = "{}.{}".format(self.fspec, self.temp_ext.lstrip("."))
            if os.path.isfile(self.temp_file):
                os.remove(self.temp_file)
            self.f = open(self.temp_file, self.mode)
        else:
            self.f = tempfile.NamedTemporaryFile(self.mode, delete=True)
            self.temp_file = self.f.name
        return self.f

    def __exit__(self, exc_type, exc_value, traceback):
        self.f.close()
        if exc_type:
            if not self.keep_temp_on_error and self.use_peer_tmp:
                os.remove(self.temp_file)
            return
        bak_file = "{}.{}".format(self.fspec, self.backup_ext.lstrip("."))
        if os.path.isfile(bak_file):
            os.remove(bak_file)
        os.rename(self.fspec, bak_file)
        os.rename(self.temp_file, self.fspec)
        if not self.keep_backup:
            os.remove(bak_file)


class VersionFileParser(ABC):
    """Base class for config file parsers."""

    def __init__(self, root_path, opts):
        self.root_path = root_path
        self.opts = opts
        self.fspec = self._find_config_file()
        self._org_version = None
        self.version = None

    def __str__(self):
        return "{}(v{})@{}".format(self.__class__.__name__, self.version, self.fspec)

    @abstractmethod
    def _find_config_file(self):
        """Return the config file path for this parser type."""

    @abstractmethod
    def parse(self, replace_version=None):
        """Read self.fspec and extract the version string."""

    @abstractmethod
    def write(self):
        """Patch the version into self.fspec and extract the version string."""

    def set_version(self, version, write=False):
        """Set a new vesion string.
        Args:
            version (str):
        """
        check_arg(version, Version)
        self.version = version
        if write:
            self.write()


class TextFileParser(VersionFileParser):
    """Read/write version numbers from plain text files.


    """

    # These `version.type`s are handled by default.
    # It is also possible to set `version.match`, `version.template`:
    pattern_map = {
        "__version__": {
            "match": r"""__version__\s*=\s*["'](\d+\.\d+\.?\d*[-+\d\w]*).*["]""",
            # "match": "__version__\s*=\s*[" '"](\d+\.\d+\.?\d*[-+\d\w]*).*[' '"]',
            # "match": "__version__\s*=\s*[" '"](\d+\.\d+\.\d+).*[' '"]',
            "template": '__version__ = "{version}"',
        },
    }

    def __init__(self, root_path, opts):
        type_defaults = self.pattern_map.get(opts["type"])
        opts.setdefault("match", type_defaults["match"])
        opts.setdefault("template", type_defaults["template"])
        super().__init__(root_path, opts)

    def _find_config_file(self):
        """Return the config file path for this parser type."""
        # default_fspec = parser.find_config_file(self.root_path)
        fspec = self.opts.get("file")
        fspec = resolve_path(self.root_path, fspec, must_exist=True)
        if not fspec:
            raise RuntimeError("Invalid `version.file`: {}".format(fspec))
        return fspec

    def parse(self, replace_version=None):
        """Read self.fspec and extract the version string."""
        version = None
        pattern = self.opts.get("match")
        pattern = re.compile(pattern)
        with open(self.fspec, "rt") as fp:
            for line in fp.readlines():
                # log_debug(line.rstrip())
                res = pattern.search(line)
                if res:
                    version = res.groups()[0]
                    # print(res)
                    # print(res.groups())

        if version is None:
            raise RuntimeError(
                "Could not match version pattern `{}` in {}".format(
                    pattern.pattern, self.fspec
                )
            )
        version = Version(version)
        self.version = version
        self._org_version = version

    def write(self):
        """Patch the version into self.fspec and extract the version string."""
        template = self.opts["template"].format(version=self.version)
        pattern = self.opts["match"]
        pattern = re.compile(pattern)

        with SafeFileWriter(self.fspec, "wt", keep_backup=False) as target:
            # target is a temporary file until __exit__, so we can open
            # the source file again here
            with open(self.fspec, "rt") as source:
                for line in source.readlines():
                    # log_debug(line.rstrip())
                    res = pattern.search(line)
                    if res:
                        # version = res.groups()[0]
                        target.write(template + "\n")
                        log_debug(
                            "Write line `{}` -> `{}`".format(
                                line.strip(), template.strip()
                            )
                        )
                    else:
                        target.write(line)
            # source is closed now. The exit handler will move the temp file
            # to the target (i.e. the original source)...
        return


class SetupCfgFileParser(VersionFileParser):
    """Read/write version numbers from setup.cfg files."""

    # def __init__(self, root_path, opts):
    #     super().__init__(root_path, opts)

    def _find_config_file(self):
        fspec = self.opts.get("file", "setup.cfg")
        fspec = resolve_path(self.root_path, fspec, must_exist=True)
        return fspec

    def parse(self, replace_version=None):
        """Read self.fspec and extract the version string."""
        cfg = ConfigParser()
        cfg.read(self.fspec)
        version = cfg.get("metadata", "version")
        if version.startswith(("attr:", "file:", "find:", "find_namespace:")):
            raise RuntimeError(
                (
                    "The version is only referenced here: '{}'. "
                    + "Use the 'file' type for the target location instead."
                ).format(version)
            )
        version = Version(version)
        self.version = version
        self._org_version = version

    def write(self):
        """Patch the version into self.fspec and extract the version string."""
        raise NotImplementedError


class PyprojectTomlParser(VersionFileParser):
    """Read/write version numbers from pyproject.toml files."""

    # def __init__(self, root_path, opts):
    #     super().__init__(root_path, opts)

    # TODO:
    # - support poetry
    # - support flit

    def _find_config_file(self, root):
        fspec = self.opts.get("file", "pyproject.toml")
        fspec = resolve_path(self.root_path, fspec, must_exist=True)
        return fspec

    def parse(self, replace_version=None):
        """Read self.fspec and extract the version string."""
        cfg = toml.load(self.fspec)
        version = cfg.get("project", {}).get("version")
        version = Version(version)
        self.version = version
        self._org_version = version

    def write(self):
        """Patch the version into self.fspec and extract the version string."""
        # TODO: assume that the section is `[tool.poetry]`!!
        raise NotImplementedError


class VersionFileManager:
    """
    Maintain a list of version locations.
    The first version parser is the 'master', additional following parsers
    are considered 'secondaries' (which will be kept in sync with master).
    """

    parser_map = {
        # "pyfile": TextFileParser,
        "pyproject": PyprojectTomlParser,
        "setup_cfg": SetupCfgFileParser,
    }

    def __init__(self, task_runner):
        self.task_runner = task_runner
        self.parsers = []
        self.org_version = None
        self.master_version = None
        self._scan_versions()

    def _scan_versions(self):
        root_path = self.task_runner.fspec
        version_opts = self.task_runner.get("version")
        if isinstance(version_opts, dict):
            version_opts = [version_opts]

        # Create a version parser for every configured location definition
        # (first is 'main', all others 'secondary')
        for vo in version_opts:
            check_arg(vo, dict)
            cfg_type = vo.get("type")
            if cfg_type is None:
                raise RuntimeError("Missing `version.type`: {}".format(vo))

            if cfg_type in TextFileParser.pattern_map:
                parser_cls = TextFileParser
            else:
                parser_cls = self.parser_map.get(cfg_type)

            if parser_cls is None:
                known_types = set(TextFileParser.pattern_map) | set(self.parser_map)
                raise RuntimeError(
                    "Invalid `version.type`: {!r} (expected: {})".format(
                        cfg_type, ", ".join(known_types)
                    )
                )

            parser = parser_cls(root_path, vo)

            # #
            # default_fspec = parser.find_config_file(root_path)
            # fspec = vo.get("file", default_fspec)
            # if not fspec:
            #     raise RuntimeError("Invalid `version.file`: {}".format(fspec))

            # fspec = resolve_path(root_path, fspec, must_exist=True)
            parser.parse()
            version = parser.version

            if self.master_version is None:
                if not version:
                    raise RuntimeError("Invalid version or no version found.")
                log_info("Parsed project version: {}".format(version))
                self.master_version = version
                self.org_version = version
            else:
                log_info("Secondary project version: {}".format(version))
            self.parsers.append(parser)
        return

    @classmethod
    def _inc_prerelease(cls, v, prefix=""):
        """Increment a version's prerelase component."""

        pre_tuple = v.prerelease
        if len(pre_tuple) == 0:  # "1.2.3" -> "1.2.3-0"
            return copy_version(v, [prefix + "0"])
        if len(pre_tuple) > 1:
            log_warning(
                "Discarding multiple prerelase identifiers in {}.".format(pre_tuple)
            )
        pre = pre_tuple[0]
        # We already have a prerelease like "1.2.3-0", "1.2.3-alpha3", "1.2.3-rc1"
        # TODO: https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
        match = re.match(r"^([a-zA-Z]*)([0-9]+)(.*)", pre)
        cur_prefix, number, rest = match.groups()
        number = int(number)
        if cur_prefix != prefix:
            log_warning(
                "Changing prerelase prefix from {!r} to {!r}.".format(
                    cur_prefix, prefix
                )
            )
        if rest:
            log_warning("Discarding prerelase sufffix {!r}.".format(rest))
        pre = "{}{}".format(prefix, number + 1)
        v_new = copy_version(v, [pre])
        return v_new

    def set_version(self, semver, write):
        check_arg(semver, Version)
        check_arg(write, bool)
        self.master_version = semver
        for parser in self.parsers:
            log_debug("Set version {}...".format(parser))
            parser.set_version(semver, write)
        return

    def reset_version(self, write):
        log_warning(
            "Reset version {} => {}...".format(self.org_version, self.master_version)
        )
        if self.master_version != self.org_version:
            self.set_version(self.org_version, write)
        return

    def bump(self, inc, write=False, prerelease_prefix="a", calc_only=False):
        check_arg(inc, str, condition=inc in INCREMENTS)
        v = self.master_version
        if inc == "major":
            # 1.2.3 -> 2.0.0
            v_next = v.next_major()
        elif inc == "minor":
            # 1.2.3 -> 1.3.0
            v_next = v.next_minor()
        elif inc == "patch":
            # 1.2.3 -> 1.2.4
            v_next = v.next_patch()
            # print("vv", v, v_next)
        elif inc == "prerelease":
            # 1.2.3   -> ERROR
            # 1.2.3-0 -> 1.2.3-1
            if not v.prerelease:
                raise RuntimeError(
                    "'prerelease' would go backwards; consider 'postrelease'"
                )
            v_next = self._inc_prerelease(v, prerelease_prefix)
        elif inc == "postrelease":
            if v.prerelease:
                # 1.2.3-0 -> 1.2.3-1
                v_next = self._inc_prerelease(v, prerelease_prefix)
            else:
                # 1.2.3   -> 1.2.4-0
                v_next = v.next_patch()
                # print(v_next)
                v_next = self._inc_prerelease(v_next, prerelease_prefix)
                # print(v_next)
        else:
            raise NotImplementedError

        log_debug("bump({}): {} -> {}".format(inc, v, v_next))
        if calc_only:
            return v_next
        self.set_version(v_next, write)
        return v_next



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os
import subprocess
from abc import ABC, abstractclassmethod, abstractmethod

from git import Repo

from .util import (
    check_arg,
    log_debug,
    log_warning,
    write,
)


GH_USER_AGENT = "Yabs/Python"


class TaskContext:
    """
    Context information that is passed by the task runner to all tasks.
    This instance is used by tasks to pass information to downstream
    tasks.
    """

    def __init__(self, args, task_runner):
        self.errors = []
        self.completed = []
        #: CLI arguments namespace object
        self._args = args
        #: (str) value of ``--inc`` argument
        #: ('major', 'minor', 'patch', 'postrelease')
        self.inc = args.inc
        #: (bool) true if ``--dry-run`` was passed
        self.dry_run = args.dry_run
        #: (str) the repo's latest tag name (before 'bump')
        self.org_tag_name = None
        #: (str) the current tag name (after 'bump')
        self.tag_name = None
        #: (dict) all files that 'pypi_release' created, e.g.
        #: ``{"sdist": <path>, "bdist_msi": <path>}``
        self.artefacts = {}
        #: (:class:`~yabs.task_runner.TaskRunner`)
        self.task_runner = task_runner
        #: (str) GitHub repo name, e.g. 'USER/PROJECT'
        self.repo = None
        #: (str) Root folder
        self.repo_path = None
        #: (:class:`git.repo.base.Repo`)
        self.repo_obj = None
        #: (str) GitHub authentication token
        self.gh_auth_token = None
        #: (:class:`semantic_version.Version`) latest version (before 'bump')
        self.org_version = None
        #: (:class:`semantic_version.Version`) current version (after 'bump')
        self.version = None
        #: (:class:`~yabs.version_manager.VersionManager`)
        self.version_manager = None

        self.initialize()
        return

    def as_dict(self):
        return vars(self)

    def initialize(self):

        if self.task_runner:
            tr = self.task_runner
            self.repo = tr.get("repo")
            self.version_manager = tr.version_manager
            self.version = self.version_manager.master_version
            self.org_version = self.version
            auth = tr.get("gh_auth")
            if isinstance(auth, str):
                self.gh_auth_token = auth
            else:  # must be a dict
                self.gh_auth_token = os.environ.get(auth["oauth_token_var"])
        else:
            self.repo_path = os.path.abspath(".")

        repo = Repo(self.repo_path)
        self.repo_obj = repo

        try:
            repo.remote().fetch(tags=True)
        except Exception as e:
            log_warning("Unable to fetch tags from git remote: {}".format(e))

        try:
            # Test if we have tags (but this is not neccessarily the latest)
            has_tags = False
            tag = repo.tags[0]
            has_tags = True
            # log_info("Latest repo tag: {}".format(tag))
        except IndexError:
            pass

        if not has_tags:
            tag = "v0.0.0"
            log_warning(
                "Repository does not seem to have tags; assuming {}".format(tag)
            )
        else:
            res = repo.git.rev_list(tags=True, max_count=1)
            tag = repo.git.describe(res, tags=True)

        self.org_tag_name = tag


class WorkflowTask(ABC):
    """
    Common base class for all yabs tasks.
    """

    _COMMON_OPTS = frozenset(("dry_run", "verbose"))
    #: (dict) define all supported arguments and their default values.
    #: This attribute must be defined by derived classes.
    DEFAULT_OPTS = None
    # CLI_COMMAND = None

    def __init__(self, opts):
        assert self.DEFAULT_OPTS is not None
        # assert self.CLI_COMMAND is not None
        #: (dict) The actual arguments, i.e. the default values merged with
        #: passed options
        self.opts = self.DEFAULT_OPTS.copy()
        self.opts.update(opts)
        check_arg(self.opts.get("dry_run"), bool)
        check_arg(self.opts.get("verbose"), int)
        #: (bool) true if `--dry-run` was passed to the CLI
        self.dry_run = self.opts.get("dry_run")
        #: (int, default=3) 0..5
        self.verbose = self.opts.get("verbose")
        known_opts = set(self.DEFAULT_OPTS).union(self._COMMON_OPTS)
        unknown_opts = set(opts.keys()).difference(known_opts)
        if unknown_opts:
            log_warning(
                "{}(): passed unknown option(s): {}".format(
                    self, ", ".join(unknown_opts)
                )
            )

    def __str__(self):
        return self.to_str({})

    def to_str(self, context):
        return "{}()".format(self.__class__.__name__)

    def _exec(self, args, quiet=None):
        """
        Args:
            args (list): array of string values that define the command line

        Returns:
            tuple (ret_code, output)
        """
        opts = self.opts
        res = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,)
        ret_code = res.returncode
        output = res.stdout.decode().strip()
        msg = "`{}` returned code {}".format(" ".join(args), ret_code)

        if ret_code != 0:
            if opts.get("ignore_errors"):
                write(msg, "warning", True, output)
                return True, None
            write(msg, "error", True, output)
        elif quiet is not True and not opts.get("silent"):
            write(msg, "debug", True, output)
        return ret_code, output

    @abstractclassmethod
    def check_task_def(cls, task_def, parser, args, yaml):  # noqa: B902 'use cls'
        """Check task definition for errors.

        This allows static pre-checks before the actual workflow starts.

        Returns:
            (str|list|bool) Error message(s)
        """
        return True

    @classmethod
    def handle_cli_command(cls, parser, args):
        """Default implementation, when run as stand-alone CLI command."""
        # Convert args namespace to option dict items:
        opts = vars(args)
        # Create task instance
        task = cls(opts)
        # The TaskRunner would maintain a `context` dict, when running a
        # sequence of workflow tasks. Here we need to set-up a simple one:
        context = TaskContext(args, None)
        # context = {
        #     "dry_run": opts["dry_run"],
        #     # "version_manager": None,
        # }
        res = task.run(context=context)
        return res

    @abstractclassmethod
    def register_cli_command(cls, subparsers, parents, run_parser):  # noqa: B902
        """Let tasks add a sub-command and/or arguments to the 'run' command."""

    @abstractmethod
    def run(self, context):
        """"""


def register_cli_commands(subparsers, parents, run_parser):
    for task_cls in WorkflowTask.__subclasses__():
        log_debug("Register {}".format(task_cls))
        task_cls.register_cli_command(subparsers, parents, run_parser)


# def register_command_handlers(handler_map):
#     for task_cls in WorkflowTask.__subclasses__():
#         handler_map[task_cls.CLI_COMMAND] = task_cls
#     return handler_map



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
from .cmd_common import WorkflowTask
from .util import check_arg, ConfigError, log_error, log_warning
from .version_manager import INCREMENTS, ORDERED_INCREMENTS


class BumpTask(WorkflowTask):
    DEFAULT_OPTS = {
        "inc": None,
        "check": True,
        "prerelease_prefix": "a",
    }

    def __init__(self, opts):
        super().__init__(opts)
        opts = self.opts
        check_arg(
            opts["inc"], str, or_none=True, condition=opts.get("inc") in INCREMENTS
        )
        check_arg(opts["check"], bool)
        check_arg(opts["prerelease_prefix"], str)

    def to_str(self, context):
        if context:
            inc = self.opts["inc"] or context.inc
            version = context.version_manager.master_version
            return "{}({!r}) => v{}".format(self.__class__.__name__, inc, version)
        return "{}({!r})".format(self.__class__.__name__, self.opts["inc"])

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        """"""
        # Additional arguments for the 'run' command
        run_parser.add_argument(
            "--inc",
            choices=["major", "minor", "patch", "postrelease"],
            default=None,
            help="bump semantic version (used as default for `bump` task's `inc` option)",
        )
        run_parser.add_argument(
            "--no-bump",
            action="store_true",
            help="run all 'bump' tasks in dry-run mode",
        )

        sp = subparsers.add_parser(
            "bump", parents=parents, help="increment current project version",
        )
        sp.add_argument(
            "inc",
            choices=["major", "minor", "patch", "postrelease"],
            # default="patch",
            help="bump semantic version by this increment",
        )
        sp.add_argument(
            "--check",
            action="store_true",
            help="test afterwards if `setup.py --version` matches",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        # We cannot check if we don't have CLI args (like in test fixtures)
        if not args:
            return True
        if args.no_bump:
            return True
        inc = task_def.get("inc", args.inc)
        if not inc:
            return "'bump' tasks require `--inc` argument or `inc` option"
        config = yaml["config"]
        max_increment = config.get("max_increment", "minor")
        max_idx = ORDERED_INCREMENTS.index(max_increment)
        inc_idx = ORDERED_INCREMENTS.index(inc)
        if inc_idx > max_idx:
            if args.force:
                log_warning(
                    "Enforcing `--inc {}` although `max_increment` option is set to '{}'".format(
                        args.inc, max_increment
                    )
                )
            else:
                return (
                    "`--inc {}` was passed, but the `max_increment` option is set to '{}'"
                    " (pass `--force` to ignore).".format(args.inc, max_increment)
                )
        return True

    def run(self, context):
        opts = self.opts
        dry_run = context.dry_run
        if context._args.no_bump:
            log_warning("`--no-bump` was passed: forcing dry-run mode for 'bump' task.")
            dry_run = True

        inc = opts.get("inc")
        if not inc:
            if context.inc:
                inc = context.inc
            else:
                raise ConfigError(
                    "Missing bump increment: either define `inc` option or pass `inc` argument."
                )
        if inc not in INCREMENTS:
            raise ConfigError(
                "Invalid `inc` option '{}' (expected {}).".format(
                    inc, ", ".join(INCREMENTS)
                )
            )

        vm = context.version_manager
        vm.bump(inc, prerelease_prefix=opts["prerelease_prefix"], write=not dry_run)
        context.version = vm.master_version

        if opts["check"] and not dry_run:
            _ret_code, real_version = self._exec(["python", "setup.py", "--version"])
            if real_version != str(vm.master_version):
                log_error(
                    "`setup.py --version` returned {} (expected {}).".format(
                        real_version, vm.master_version
                    )
                )
                return False

        return True



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os

from git import Repo
from git.exc import GitCommandError

from .cmd_common import WorkflowTask
from .util import check_arg, log_response


class PushTask(WorkflowTask):
    DEFAULT_OPTS = {
        "target": "",  # E.g. 'upstream'
        "tags": False,  # Also push tags
        # "useFollowTags": False,  # Use `--folow-tags` instead of `&& push --tags`
    }

    def __init__(self, opts):
        super().__init__(opts)

        opts = self.opts
        check_arg(opts["target"], str)
        check_arg(opts["tags"], bool)
        # check_arg(opts["useFollowTags"], bool)

    # def to_str(self, context):
    #     add = self.opts["add"] or self.opts["add_known"]
    #     return "{}(add: {}, '{}')".format(
    #         self.__class__.__name__, add, self.opts["message"]
    #     )

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        """"""
        sp = subparsers.add_parser(
            "push",
            parents=parents,
            help="increment current 'patch' version (add `--minor` or `--major`)",
        )
        sp.add_argument(
            "--tags", action="store_true", help="push missing but relevant tags",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        opts = self.opts
        target = self.opts["target"]
        repo_path = os.path.abspath(".")
        repo = Repo(repo_path)
        git = repo.git

        try:
            if target:
                res = git.push(
                    opts["target"],
                    follow_tags=opts["tags"],
                    dry_run=self.dry_run,
                    verbose=self.verbose >= 4,
                )
            else:
                res = git.push(
                    follow_tags=opts["tags"],
                    dry_run=self.dry_run,
                    verbose=self.verbose >= 4,
                )
            log_response("git push", res, "info", self.dry_run)
        except GitCommandError as e:
            log_response("git push", "{}".format(e), "error", self.dry_run)
            return False
        return True



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import subprocess

from .cmd_common import WorkflowTask
from .util import check_arg, log_dry, log_response


class ExecTask(WorkflowTask):
    DEFAULT_OPTS = {
        "args": [],
        "dry_run_args": None,
        "always": False,
        "silent": False,
        "ignore_errors": False,
        "timeout": None,
    }

    def __init__(self, opts):
        super().__init__(opts)

        opts = self.opts
        check_arg(opts["args"], (list, tuple))
        check_arg(opts["dry_run_args"], (list, tuple), or_none=True)
        check_arg(opts["silent"], bool)
        check_arg(opts["always"], bool)
        check_arg(opts["ignore_errors"], bool)
        check_arg(opts["timeout"], int, or_none=True)
        if opts["dry_run_args"] and opts["always"]:
            raise RuntimeError("`dry_run_args` and `always` are mutually exclusive")

    def to_str(self, context):
        opts = self.opts
        if self.dry_run and opts["dry_run_args"] is not None:
            args = opts["dry_run_args"]
        else:
            args = opts["args"]
        return "{}(`{}`)".format(self.__class__.__name__, " ".join(args))

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        """"""
        sp = subparsers.add_parser(
            "exec", parents=parents, help="execute shell command",
        )
        sp.add_argument(
            "args", nargs="+", help="shell command and arguments",
        )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        opts = self.opts

        args = opts["args"]
        if self.dry_run:
            if not opts["always"] and opts["dry_run_args"] is None:
                log_dry("Execute {}".format(" ".join(opts["args"])))
                return True
            if opts["dry_run_args"] is not None:
                args = opts["dry_run_args"]

        res = subprocess.run(
            args,
            timeout=opts["timeout"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=False,
        )
        ret_code = res.returncode
        output = res.stdout.decode()
        msg = "{} returned code {}".format(self.to_str(context), ret_code)

        if ret_code != 0:
            if opts["ignore_errors"]:
                log_response(msg, output, "warning", False)
                return True
            log_response(msg, output, "error", False)
        elif not opts["silent"]:
            log_response(msg, output, "info", False)

        return ret_code == 0



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os

from github import Github, GithubObject

from .cmd_common import WorkflowTask, GH_USER_AGENT
from .util import (
    ConfigError,
    check_arg,
    log_dry,
    log_error,
    log_info,
    log_ok,
    log_warning,
)


class GithubReleaseTask(WorkflowTask):
    DEFAULT_OPTS = {
        "gh_auth": None,  # use `config.gh_auth`
        "draft": False,  # Use `--gh-draft` to override
        "message": "Released {version}\n"
        + "[Commit details](https://github.com/{repo}/compare/{org_tag_name}...{tag_name}).",
        "name": "v{version}",
        "prerelease": None,  # None: guess from version number; Use `--gh-pre` to override
        "repo": None,  # `owner/repo`, defaults to yaml setting
        "tag": None,
        "target_commitish": None,
        "upload": None,
    }

    def __init__(self, opts):
        super().__init__(opts)

        opts = self.opts
        check_arg(opts["gh_auth"], (dict, str), or_none=True)
        check_arg(opts["repo"], str, or_none=True)
        check_arg(opts["name"], str)
        check_arg(opts["message"], str)
        check_arg(opts["draft"], bool)
        check_arg(opts["prerelease"], bool, or_none=True)
        check_arg(opts["upload"], list, or_none=True)
        check_arg(opts["target_commitish"], str, or_none=True)
        check_arg(opts["tag"], str, or_none=True)

        upload = opts["upload"]
        if upload:
            unknown = set(upload).difference(set(("sdist", "bdist_wheel", "bdist_msi")))
            if unknown:
                raise RuntimeError(
                    "Unknown upload target(s): {}".format(", ".join(unknown))
                )

    # def to_str(self, context):
    #     add = self.opts["add"] or self.opts["add_known"]
    #     return "{}(add: {}, '{}')".format(
    #         self.__class__.__name__, add, self.opts["message"]
    #     )

    @classmethod
    def register_cli_command(cls, subparsers, parents, run_parser):
        # Additional arguments for the 'run' command
        run_parser.add_argument(
            "--gh-draft",
            action="store_true",
            help="tell the gh_release task to create a draft-release",
        )
        run_parser.add_argument(
            "--gh-pre",
            action="store_true",
            help="tell the gh_release task to create a pre-release",
        )
        # New sub-command
        sp = subparsers.add_parser(
            "gh-release", parents=parents, help="create a release on GitHub",
        )
        # sp.add_argument(
        #     "--draft", action="store_true", help="",
        # )
        # sp.add_argument(
        #     "--opts", help="YAML file with conversion options",
        # )
        sp.set_defaults(command=cls.handle_cli_command)

    @classmethod
    def check_task_def(cls, task_def, parser, args, yaml):
        return True

    def run(self, context):
        opts = self.opts

        if context._args.no_release:
            log_warning("`--no-release` was passed: skipping 'gh_release' task.")
            return True

        ok = True
        # GitHub access token
        auth = opts["gh_auth"]
        if auth:
            if isinstance(auth, dict):
                token = os.environ.get(auth["oauth_token_var"])
            else:
                token = auth
        else:
            token = context.gh_auth_token
        gh = Github(token, user_agent=GH_USER_AGENT)

        repo_name = opts.get("repo") or context.repo
        if not repo_name or "/" not in repo_name:
            raise ConfigError("Missing repo name (expected `GH-USER/PROJECT`).")

        repo = gh.get_repo(repo_name, lazy=False)
        if not repo:
            raise ConfigError("Could not open repo '{}'.".format(repo_name))

        # collaborators = repo.get_collaborators("all")
        # for collab in collaborators:
        #     print(collab)

        # releases = repo.get_releases()
        # for release in releases:
        #     print(release)

        # branches = repo.get_branches()
        # for branch in branches:
        #     print(branch)

        tag_name = context.tag_name
        if not tag_name:
            log_error("Please run 'tag' task first, to create a new tag.")
            return False

        gh_tag = None
        for t in repo.get_tags():
            # print("GH tag", t)
            if t.name == tag_name:
                gh_tag = t

        if self.dry_run:
            log_dry("create_git_release({})".format(gh_tag))
            return True

        if not gh_tag:
            log_error(
                "Could not find tag '{}' on Github (did you run 'bump' and 'tag' tasks first?)".format(
                    tag_name
                )
            )
            return False

        target_commitish = opts["target_commitish"] or GithubObject.NotSet

        draft = bool(context._args.gh_draft or opts["draft"])

        if context._args.gh_pre:
            prerelease = True
        else:
            prerelease = opts["prerelease"]
            if prerelease is None:
                prerelease = "-" in tag_name or "+" in tag_name
                log_info(
                    "Tag '{}': assuming prerelease={}".format(tag_name, prerelease)
                )

        name = opts["name"].format(**vars(context))
        message = opts["message"].format(**vars(context))

        # gh_tag = repo.get_git_tag()
        gh_release = repo.create_git_release(
            tag=tag_name,
            name=name,
            message=message,
            draft=draft,
            prerelease=prerelease,
            target_commitish=target_commitish,
        )

        if opts["upload"]:
            for target in opts["upload"]:
                path = context.artefacts.get(target)
                if not path or not os.path.isfile(path):
                    raise RuntimeError("Artefact not found: {}".format(path))
                gh_asset = gh_release.upload_asset(
                    str(path),
                    label="",
                    # content_type=NotSet,
                    # name=NotSet,
                )
                log_ok("Upload asset {}".format(gh_asset))

        return ok



# -*- coding: utf-8 -*-
# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
# Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
"""
"""
import os
import time

import yaml

from snazzy import emoji

from .cmd_bump import BumpTask
from .cmd_check import CheckTask
from .cmd_commit import CommitTask
from .cmd_common import TaskContext
from .cmd_exec import ExecTask
from .cmd_gh_release import GithubReleaseTask
from .cmd_push import PushTask
from .cmd_pypi_release import PypiReleaseTask
from .cmd_tag import TagTask
from .util import (
    NO_DEFAULT,
    ConfigError,
    check_arg,
    format_elap,
    log_debug,
    log_error,
    log_ok,
    log_warning,
    resolve_path,
)
from .version_manager import VersionFileManager


class TaskRunner:
    """"""

    # TODO: automate this map creation:
    handler_map = {
        "bump": BumpTask,
        "check": CheckTask,
        "commit": CommitTask,
        "exec": ExecTask,
        "github_release": GithubReleaseTask,
        "push": PushTask,
        "pypi_release": PypiReleaseTask,
        "tag": TagTask,
    }

    def __init__(self, fspec, parser=None, args=None):
        self.fspec = fspec
        self.parser = parser
        self.args = args
        self.all = None
        self.config = None
        self.tasks = None
        self.version_manager = None
        self._load()
        self._check_config(parser, args)
        # register_command_handlers(self.handler_map)

    def get(self, key, default=NO_DEFAULT):
        try:
            return self.config[key]
        except KeyError:
            if default is NO_DEFAULT:
                raise
        return default

    def make_abs_path(self, key, default=None):
        return self.config.get(key, default)

    def _load(self):
        with open(self.fspec, "rt") as f:
            try:
                res = yaml.safe_load(f)
            except yaml.parser.ParserError as e:
                raise RuntimeError("Could not parse YAML: {}".format(e)) from None

        if not isinstance(res, dict) or not res.get("file_version", "").startswith(
            "yabs#"
        ):
            raise ConfigError("Not a `yabs` file (missing 'yabs#VERSION' tag).")
        self.all = res
        self.config = res["config"]
        check_arg(self.config, dict)
        self.tasks = res["tasks"]
        check_arg(self.tasks, list)

        self.version_manager = VersionFileManager(self)
        return

    def _check_config(self, parser, args):
        """Early option and command line syntax checks.

        This is done after reading the YAML and parsing the command line,
        but before the workflow starts.
        Note that `parser` and `args` may be None, when the TaskRunner was
        created by a srcipt (or test fixture).
        """
        errors = []

        for task_def in self.tasks:
            task_type = task_def["task"]
            task_cls = TaskRunner.handler_map.get(task_type)
            if task_cls is None:
                errors.append(
                    "Invalid task definition: '{}': {}".format(task_type, task_def)
                )
                continue
            res = task_cls.check_task_def(task_def, parser, args, self.all)
            if res in (None, True):
                continue
            if res is False:
                res = "{}({}): check failed.".format(task_cls, task_def)
            if isinstance(res, str):
                res = [res]
            check_arg(res, (list, tuple), or_none=True)
            errors.extend(res)

        if errors:
            raise ConfigError("ConfigError:\n  - {}".format("\n  - ".join(errors)))
        return True

    def run(self):
        context = TaskContext(self.args, self)
        ok = True
        start = time.monotonic()
        for task_def in self.tasks:
            task_def = task_def.copy()
            # log_info(task_def)
            task_type = task_def.pop("task")
            task_cls = TaskRunner.handler_map.get(task_type)
            if not task_cls:
                raise ConfigError(
                    "Invalid task type: {} (expected {})".format(
                        task_type, ", ".join(self.handler_map.keys())
                    )
                )
            # TODO: task_def can force - but not prevent - dry-run:
            task_def["dry_run"] = self.args.dry_run
            task_def["verbose"] = self.args.verbose
            task = task_cls(task_def)
            task_str = task.to_str(context)
            log_debug("Running {}: {}...".format(task_str, task.opts))
            res = task.run(context)
            task_str = task.to_str(context)  # __str__ may have changed
            if res:
                log_ok("{}".format(task_str))
            else:
                log_error("{}".format(task_str))
                context.errors.append(task_str)
                ok = False
                # if not args.force_continue:
                break

        elap = time.monotonic() - start
        if ok:
            log_ok(
                "Workflow finished successfully in {}{}".format(
                    format_elap(elap), emoji(" ✨ 🍰 ✨")
                )
            )
        else:
            msg = "Workflow failed in {}{}".format(format_elap(elap), emoji(" 💥 💔 💥"))
            log_error(msg)
            context.errors.append(msg)

        if self.args.dry_run:
            log_warning(
                "Dry-Run mode: No bits were harmed during the making of this release."
            )
        return ok

    # def find_project_root(self, fspec="."):
    #     """Return the nearest parent path that conatains setup.py."""
    #     path = Path(fspec).absolute()
    #     while path.parents:
    #         log_info("Searching {}...".format(path))
    #         if (path / "setup.py").is_file():
    #             return path
    #         path = path.parent
    #     return None


def handle_run_command(parser, args):
    fspec = resolve_path(os.getcwd(), args.workflow, must_exist=True)
    tm = TaskRunner(fspec, parser, args)
    res = tm.run()
    return res
